# Import argparse
import argparse
from textwrap import dedent
import numpy as np

# Import Modules
from . import core
from . import phonon

# Create the wrapper funstions 

def generate_input_func(args):
    '''
    Wrapper function for the generate_input function

    parameters
    ----------
    filename : str
    config_file : file

    Returns
    -------
    INCAR : file
    '''
    core.generate_input(
        filename=args.filename,
        calculation=args.config_file,
            )

def calculate_kpoints_func(args):
    '''
    Wrapper function for the calculate_kpoints function

    parameters
    ----------
    filename : str

    Returns
    -------
    KPOINT mesh : int 
    '''
    core.calculate_kpoints(
            filename=args.filename,
            )

def generate_kpoints_func(args):
    '''
    Wrapper function for the generate_kpoints function 

    parameters
    ----------
    mesh : int 

    Returns
    -------
    KPOINTS : file
    '''
    core.generate_kpoints(
            mesh=args.mesh,
            )

def generate_supercell_func(args):
    '''
    Wrapper function for the generate_supercell function

    parameters
    ----------
    expansion : list

    Returns
    -------
    POSCAR : file 
    .xyz : file
    '''
    core.generate_supercell(
            expansion=np.array(args.expansion, dtype=int),
            )

def generate_job_func(args):
    '''
    Wrapper function for the generate_job function 

    parameters
    ----------
    title : str 
    cores : int 
    vasp_type : str 

    Returns
    -------
    job.sh : file
    '''
    core.generate_job(
            title=args.title,
            cores=args.cores,
            vasp_type=args.vasp_type,
            )

def start_up_func(args):
    '''
    Wrapper function for the start_up function

    parameters
    ----------
    None

    Returns
    -------
    configuration files : file
    '''
    core.create_input_configurations()

def dope_structure_func(args):
    '''
    Wrapper function for the dope_structure function

    parameters
    ----------
    filename : str 
    dopant : str 
    replace : str 
    instances : int 
    
    Returns
    -------
    POSCARs : file 
    '''
    core.dope_structure(
        filename=args.filename,
        dopant=args.dopant,
        replace=args.replace,
        instances=args.instances,
        )

def generate_defect_func(args): 
    '''
    Wrapper function for the generate_defect function

    parameters
    ----------
    filename : str 
    site : str 
    instances : int 

    Returns
    -------
    POSCAR : file 
    '''
    core.generate_defect(
            filename=args.filename,
            site=args.site,
            instances=args.instances,
            )

def asymmetric_unit_func(args):
    '''
    Wrapper function for the asymmetric_unit function

    parameters
    ----------
    filename : str 
    atom : str 
    bond_max : float 

    Returns
    -------
    POSCAR : file 
    '''
    core.asymmetric_unit(
            filename=args.filename,
            atom=args.atom,
            bond_max=args.bond_max,
            )

def Ewald_func(args):
    '''
    Wrapper function for the Ewald function

    parameters
    ----------
    qm_region : str 
    filename : str 
    charges : str 
    atom : str 
    n : list 
    r_cut : float 
    expansion : list 

    Returns 
    ------- 
    Ewald.out : file 
    '''
    core.Ewald(
            qm_region=args.qm_region,
            filename=args.filename,
            charges=args.charges,
            atom=args.atom,
            n=args.n,
            r_cut=args.r_cut,
            expansion=args.expansion,
            bound=0.0,
            verbose=args.verbose,
            )

def potential_plot_func(args):
    '''
    Wrapper function for the potential_plot function

    parameters
    ----------
    filename : str 
    xy_grid : int 
    n_points : int 
    n_contours : int

    Returns
    -------
    None
    '''
    core.potential_plot(
            filename=args.filename,
            xy_grid=args.xy_grid, 
            n_points=args.n_points, 
            n_contours=args.n_contours,
            )

def convert_cif_func(args):
    '''
    Wrapper function for the convert_cif function

    parameters
    ----------
    filename : str 

    Returns
    -------
    POSCAR : file 
    '''
    core.convert_cif(
            filename=args.filename,
            )

def qm_region_func(args):
    '''
    Wrapper function for the qm_region function

    parameters
    ----------
    filename : str 
    atom : str 
    r_cut : float 
    expansion : list 

    Returns
    -------
    POSCAR : file 
    '''
    core.qm_region(
            filename=args.filename,
            atom=args.atom,
            )

def phonon_calc(args):
    '''
    wrapper function for phonon calculation

    parameters
    ----------
    supercell : list
    mesh : list
    encut : int
    cores : int
    bec : bool

    Returns
    -------
    None
    '''
    supercell = args.supercell
    mesh = args.mesh
    encut = args.encut
    cores = args.cores
    bec = args.bec

# Check if the config file exists
    if not phonon.check_config():
        raise ValueError('No config file found, please run the start_up command')

# Generate the input files
    # core.generate_input('POSCAR', 'phonon')
    core.generate_kpoints(mesh=mesh)

    with open('INCAR', 'a') as f:
        f.write('ENCUT = {}\n'.format(str(encut)))
        if bec:
            f.write('LEPSILON = .TRUE.')

# Create displacements
    phonon.create_displacements(supercell)

# get the number of displacements
    disp = phonon.disp_files()

# Generate the job file
    job = phonon.phonopy_submit(cores, mesh)
    job.write_submit(disp)


def read_args(arg_list=None):
    '''Reads the command line arguments'''
    parser = argparse.ArgumentParser(
            prog='vasp_suite',
            description=dedent(
                '''
                ---------------------------------------------------
                                                  
                      A suite of tools for VASP calculations      

                ---------------------------------------------------
                
                Available programmes:
                    vasp_suite generate_input ...
                    vasp_suite generate_job ...
                    vasp_suite calculate_kpoints ...
                    vasp_suite generate_kpoints ...
                    vasp_suite generate_supercell ...
                    vasp_suite convert_cif ...
                    vasp_suite start_up ...
                    vasp_suite dope_structure ...
                    vasp_suite generate_defect ...
                    vasp_suite molecular_qm ...
                    vasp_suite plot_potential ...
                    vasp_suite qm_region ...
                    vasp_suite phonon_calc ...

                ##############################
                ## Ewald code now found in: ##
                ##        env_suite         ##
                ##############################
                '''
                ),
            epilog=dedent(
                '''
                To display options for a specific programme, use vasp_suite <programme> -h
                '''
                ),
            formatter_class=argparse.RawDescriptionHelpFormatter
            )

    # Subparsers
    subparsers = parser.add_subparsers(dest='prog')

    gen_inp = subparsers.add_parser(
            'generate_input',
            help='Generate input files for VASP calculations',
            description=dedent(
                '''
                Generation of INCAR and POTCAR files for VASP calculations.
                '''
                ),
            formatter_class=argparse.RawDescriptionHelpFormatter
            )

    gen_inp.set_defaults(func=generate_input_func)

    gen_inp.add_argument(
            'config_file',
            help=dedent(
                '''
                The configuration file for the input generation.
                
                Example:
                Inside ~/.vasp_suite_templates '.ini' configuration files are 
                stored. To perform a relaxation caluclation using the relaxation.ini 
                template, use the following command:

                vasp_suite generate_input relaxation
                '''
                )
            )

    gen_inp.add_argument(
            '--filename', '-f',
            help=dedent(
                '''
                The name of the structure file, default is POSCAR
                '''
                ),
            default='POSCAR'
            )

    gen_job = subparsers.add_parser(
        'generate_job',
        help='Generate job submission files for VASP calculations',
        formatter_class=argparse.RawDescriptionHelpFormatter
        )

    gen_job.set_defaults(func=generate_job_func)

    gen_job.add_argument(
        'title',
        help='The title of the job'
        )

    gen_job.add_argument(
        'cores',
        help='The number of cores to use',
        type=int,
        )

    gen_job.add_argument(
        '--vasp_type',
        help=dedent(
            '''
            The VASP programme you widh to use:
                - vasp_std
                - vasp_gam
            '''
            ),
        default='vasp_gam'
        )

    calc_kpoints = subparsers.add_parser(
        'calculate_kpoints',
        help='Calculate the possible kpoint meshes for a given structure',
        formatter_class=argparse.RawDescriptionHelpFormatter
        )

    calc_kpoints.set_defaults(func=calculate_kpoints_func)

    calc_kpoints.add_argument(
            '--filename', '-f',
            help=dedent(
                '''
                The name of the structure file, default is POSCAR
                '''
                ),
            default='POSCAR'
            )

    gen_kpoints = subparsers.add_parser(
        'generate_kpoints',
        help='Generate the kpoints file for a given structure',
        formatter_class=argparse.RawDescriptionHelpFormatter
        )

    gen_kpoints.set_defaults(func=generate_kpoints_func)

    gen_kpoints.add_argument(
        'mesh',
        nargs=3,
        help=dedent(
            '''
            The mesh to use for the kpoints file.
            command line arguments are written in the form:
            a b c
            '''
            )
        )

    gen_supercell = subparsers.add_parser(
        'generate_supercell',
        help='Generate a supercell from a given structure',
        formatter_class=argparse.RawDescriptionHelpFormatter
        )

    gen_supercell.set_defaults(func=generate_supercell_func)

    gen_supercell.add_argument(
        'expansion',
        nargs=3,
        help=dedent(
            '''
            The expansion vector for the supercell, a b c
            '''
            ),
        )

    start_up = subparsers.add_parser(
        'set_up',
        help='Generate the configuration files for the suite',
        formatter_class=argparse.RawDescriptionHelpFormatter
        )

    start_up.set_defaults(func=start_up_func)

    dope_struct = subparsers.add_parser(
        'dope_structure',
        help='Dope a structure with a given element',
        formatter_class=argparse.RawDescriptionHelpFormatter
        )

    dope_struct.set_defaults(func=dope_structure_func)

    dope_struct.add_argument(
        '--filename', '-f',
        help='The name of the structure file',
        default='POSCAR',
        )

    dope_struct.add_argument(
            'dopant',
            help='The element to dope the structure with',
            )

    dope_struct.add_argument(
            'replace',
            help='The element to replace',
            ) 

    dope_struct.add_argument(
            '--instances',
            help='The number of instances of the dopant to add',
            type=int,
            default=1,
            )

    gen_defect = subparsers.add_parser(
        'generate_defect',
        help='Generate a defect structure',
        formatter_class=argparse.RawDescriptionHelpFormatter
        )

    gen_defect.set_defaults(func=generate_defect_func)

    gen_defect.add_argument(
            '--filename', '-f',
            help='The name of the structure file',
            default='POSCAR',
            )

    gen_defect.add_argument(
            'site',
            help='The name of the atom to remove',
            )

    gen_defect.add_argument(
            '--instances',
            help='The number of instances of defect',
            type=int,
            default=1,
            )

    asym = subparsers.add_parser(
        'molecular_qm',
        help='Generate the asymmetric unit of a structure',
        formatter_class=argparse.RawDescriptionHelpFormatter
        )

    asym.set_defaults(func=asymmetric_unit_func)

    asym.add_argument(
            '--filename', '-f',
            help='The name of the structure file',
            default='POSCAR',
            )

    asym.add_argument(
            'atom',
            help='The name of the spin centre in the molecular crystal',
            )

    asym.add_argument(
            '--bond_max', '-b',
            help='The maximum bond length/ Å',
            type=float,
            default=2.6,
            )

    ewald = subparsers.add_parser(
        'ewald',
        help='Calculate the ewald potential and fit paramaeter charges to the potential for electronic structure calculations',
        formatter_class=argparse.RawDescriptionHelpFormatter
        )

    ewald.set_defaults(func=Ewald_func)

    ewald.add_argument(
            '--filename', '-f',
            help='The name of the structure file',
            default='POSCAR',
            )

    ewald.add_argument(
            'qm_region',
            help='The name of the qm region, (.xyz file created using asymmetric unit)',
            )

    ewald.add_argument(
            'charges',
            help='The name of the file containing the charges',
            )

    ewald.add_argument(
            'atom',
            help='The name of the spin centre in the molecular crystal <Dy1>',
            )

    ewald.add_argument(
            '--n',
            help='The summation expansion',
            default= [3, 3, 3],
            nargs=3,
            type=int,
            )

    ewald.add_argument(
            'r_cut',
            help='The cutoff radius for the sphere region',
            type=float,
            )

    ewald.add_argument(
            'expansion',
            help='The expansion vector for the supercell, a b c',
            nargs=3,
            type=int,
            )

    ewald.add_argument(
            '--verbose', '-v',
            help='The level of verbosity in the fitting process\n0: No output\n1: Optimility and cost function\n2: Full output',
            default=1,
            type=int,
            )

    pot = subparsers.add_parser(
            'plot_potential',
            help='Plot the potential in the xy plane for a given mesh grid',
            formatter_class=argparse.RawDescriptionHelpFormatter
            )

    pot.set_defaults(func=potential_plot_func)

    pot.add_argument(
            '--filename', '-f',
            help='The name of the data file. delfault is ewald.out\n The layout of the file should be:\n x y z charge',
            default='ewald.out',
            )

    pot.add_argument(
            'xy_grid',
            help='The length in angstrom of the mesh grid in the xy plane. from -xy_grid to xy_grid',
            type=int,
            )

    pot.add_argument(
            '--n_points', '-p',
            help='The number of points in the mesh grid',
            type=int,
            default=50,
            )

    pot.add_argument(
            '--n_contours', '-c',
            help='The number of contours to plot',
            type=int,
            default=100,
            )

    cif = subparsers.add_parser(
            'convert_cif',
            help='Generate a cif file from a structure file',
            formatter_class=argparse.RawDescriptionHelpFormatter
            ) 

    cif.set_defaults(func=convert_cif_func) 

    cif.add_argument(
            'filename',
            help='The name of the .CIF file',
            )

    qm = subparsers.add_parser(
            'qm_region',
            help='Generate a .xyz file containing the qm region for non-molecular crystals',
            formatter_class=argparse.RawDescriptionHelpFormatter
            )

    qm.set_defaults(func=qm_region_func)

    qm.add_argument(
            '--filename', '-f',
            help='The name of the structure file',
            default='POSCAR',
            )

    qm.add_argument(
            'atom',
            help='The name of the spin centre in the crystal, <Dy1>',
            )

    phonon = subparsers.add_parser(
            'phonon_calc',
            help='Generate a phonopy input files and submit files',
            formatter_class=argparse.RawDescriptionHelpFormatter
            )

    phonon.set_defaults(func=phonon_calc)

    phonon.add_argument(
            '-mesh',
            help='The mesh to use for the phonon calculation',
            nargs=3,
            type=int,
            )

    phonon.add_argument(
            '-cores',
            help='The number of cores to use for the phonon calculation',
            type=int,
            )

    phonon.add_argument(
            '-encut',
            help='The energy cutoff for the phonon calculation',
            type=int,
            )

    phonon.add_argument(
            '--supercell', '-s',
            help='The supercell to use for the phonon calculation',
            nargs=3,
            type=int,
            default=[1, 1, 1],
            )

    phonon.add_argument(
            '--bec',
            help='Calculate the Born effective charges',
            action='store_true',
            )

    # Parse the ArgumentParser
    parser.set_defaults(func=lambda args: parser.print_help())
    args = parser.parse_known_args(arg_list)

    # Select programme 
    if args in ['generate_input, generate_job, calculate_kpoints, generate_kpoints, generate_supercell, start_up, dope_structure, generate_defect, asymmetric_unit, ewald, plot_potential, convert_cif']:
        args.func(args)
    else:
        args = parser.parse_args(arg_list)
        args.func(args)

def main():
    read_args()



'''
A programme which holds the core functions of vasp_suite.
'''

# Imports 
import numpy as np 
import os
from . import structure 
from . import input 
from . import submission
from . import ewald
import socket
# Functions 

def generate_input(filename: str, calculation: str) -> None:
    '''
    Generates Input files for VASP Calculations.
    '''
    input_files = input.InputFileGenerator(filename, calculation)
    input_files._INCAR()
    input_files._POTCAR()
    return None

def generate_job(title, cores, vasp_type):
    '''
    Generates a submission script for VASP calculations.
    '''
    job = submission.SubmissionWriter(title, cores, vasp_type)
    if job.hostname == 'csf3':
        job.submission_csf3() 
    elif job.hostname == 'csf4':
        job.submission_csf4() 
    else: 
        exit()

def generate_supercell(expansion: np.ndarray) -> None: 
    '''
    Generates a supercell from a given expansion matrix.
    '''
    _s = structure.Structure('POSCAR')
    _s.Vasp_reader()
    _s.Supercell(expansion)
    _s.supercell_reorder()
    _s.write_poscar()
    return None

def dope_structure(filename: str, dopant: str, replace: str, instances: int) -> None:
    '''
    Dopes a structure with a given dopant and generates all possible structures.
    '''
    _d = structure.DOPE(filename, dopant, replace, instances)
    _d.Generate_structure()
    _d.write_poscars()
    return None

def generate_defect(filename: str, site: str, instances: int) -> None:
    '''
    Generates a defect structure from a given structure.
    '''
    _d = structure.DOPE(filename, 'D', site, instances)
    _d.Generate_structure()
    _d.Create_defect()
    _d.write_poscars()
    return None

def asymmetric_unit(filename: str, atom: str, bond_max: float) -> None:
    '''
    Generates the asymmetric unit of a structure.
    '''
    _a = structure.Asymmetric_unit(filename, atom, bond_max)
    _a.find_unit()


def calculate_kpoints(filename: str) -> None:
    '''
    Calculates possible kpoint meshes for a given structure.
    '''
    _s = structure.Structure(filename) 
    _s.Vasp_reader() 
    mesh, density = _s.Calculate_mesh()
    for i in range(len(mesh)):
        print(f'Mesh: {mesh[i][0]} {mesh[i][1]} {mesh[i][2]} Density: {density[i]}')
    return None

def generate_kpoints(mesh: list) -> None:
    with open('KPOINTS', 'w') as f:
        f.write(f'''Regular {mesh[0]} {mesh[1]} {mesh[2]} gamma centred mesh \n''')
        f.write('0\nGamma\n')
        f.write(f'{mesh[0]} {mesh[1]} {mesh[2]}\n')
        f.write('0 0 0')

def convert_cif(filename: str) -> None:
    '''
    Converts a cif file to a POSCAR file.
    '''
    _s = structure.Structure(filename)
    _s.CIF_reader()
    _s.write_poscar()

def qm_region(filename: str, atom: str):
    '''
    Generates a qm region for a non-molecular crystal
    '''
    _s = structure.Structure(filename)
    _s.Vasp_reader()
    _s._find_index(atom)
    _s.Const_shift(_s.vector)
    _s.Supercell([2, 2, 2])
    _s._type = 'Cartesian'
    _s.Const_shift(-np.array([0.5, 0.5, 0.5]))
    _s.Cart_coords()
    _s._build_region()
    _s.write_xyz()

    

def Ewald(
        filename: str,
        qm_region: str,
        charges: str,
        atom: str,
        n: list,
        expansion: np.ndarray,
        r_cut: float,
        bound: float,
        verbose: int,
        ) -> None:
    '''
    Generates a xfield input file for molcas calulations using the Ewald summation.
    '''
    if 'csf' in socket.gethostname() and 'login' in socket.gethostname():
        raise Warning('DO NOT run ewald on the CSF login node!!!')
    _e = ewald.Ewald(
            qm_region,
            filename,
            charges,
            atom,
            n,
            r_cut,
            expansion,
            bound,
            verbose
            )
    _e.Calculate_Ewald()
    _e.Create_Zones()
    _e.Fitting()

def potential_plot(
        filename: str,
        xy_grid: int,
        n_points: int,
        n_contours: int,
        ) -> None:
    '''
    Generates a potential plot from a molcas xfield output file.
    '''
    pot = ewald.Potential(filename, xy_grid, n_points, n_contours)
    pot.Plot()

def create_input_configurations() -> None:
    '''
    Creates input configurations for VASP calculations.
    '''
    if not os.path.exists(os.path.expanduser('~/.vasp_suite_configs')):
        os.mkdir(os.path.expanduser('~/.vasp_suite_configs'))

    cwd = os.getcwd()
    os.chdir(os.path.expanduser('~/.vasp_suite_configs'))
    with open('relaxation.ini', 'w') as f:
        f.write('''[Relaxation]
prec = ACCURATE
lreal = .FALSE.
lasph = .TRUE.
ismear = 0
sigma = 0.1
nelm = 100
nelmin = 4
ncore = 4
ediff = 1e-08
ediffg = -0.01
ibrion = 2
nsw = 100
isif = 4
potim = 0.5
lwave = .FALSE.
lcharg = .FALSE.
lorbit = 11
gga = PE
ivdw = 11
''')
    with open('scf.ini', 'w') as f:
        f.write('''[SCF] 
                prec = ACCURATE
lreal = .FALSE.
lasph = .TRUE.
ismear = 0
sigma = 0.1
nelm = 100
nelmin = 4
ncore = 4
ediff = 1e-08
ediffg = -0.01
ibrion = -1
isif = 4
potim = 0.5
lwave = .FALSE.
lcharg = .FALSE.
lorbit = 11
gga = PE
ivdw = 11
''')
    with open('phonon.ini', 'w') as f:
        f.write('''[Phonon]
ALGO = Normal
EDIFF = 1E-8
GGA = PE
ISIF = 2
ISMEAR = 0
LASPH = .TRUE.
LCHARG = .FALSE.
LREAL = .FLASE.
LWAVE = .FALSE.
NSW = 0
PREC = Accurate
SIGMA = 0.01
''')
    os.chdir(cwd)



'''
A module for the Structure class. Structure will perform operations structure data files.
'''
 # Imports 
import re
import os
import numpy as np
import scipy as sp
from functools import wraps
from time import perf_counter
from ase import io


def timer(func):
    '''
    A decorator function to time functions.
    '''
    @wraps(func)
    def wrapper(*args, **kwargs):
        start = perf_counter()
        print(f'Running {func.__name__}...')
        func(*args, *kwargs)
        end = perf_counter()
        print(f"Time elapsed: {end - start}")
    return wrapper


class Structure():
    '''
    A class of functions for manipulating structure data files.
    '''

    def __init__(self, filename: str):
        '''
        Initializes the Structure class.
        '''
        self.filename = filename

    def Vasp_reader(self):
        '''
        Reads and formats a .vasp, POSCAR or CONTCAR file.
        '''
        with open(self.filename) as f:
            data = f.readlines()
        data = [x.strip().split() for x in data]

        self.name: str = data[0][0]
        self.scale: float = float(data[1][0])
        self.av: np.ndarray = np.array(data[2:5], dtype=float)
        self.atoms: list = data[5]
        self.natoms: list = [int(x) for x in data[6]]
        self._N: int = sum(self.natoms)
        self._type: str = data[7][0].lower()
        self.coords: np.ndarray = np.array(data[8:8+self._N], dtype=float)

        # Calculate reciprocal lattice vectors
        a1, a2, a3 = self.av
        b1 = 2 * np.pi * np.cross(a2, a3) / (a1@np.cross(a2, a3))
        b2 = 2 * np.pi * np.cross(a3, a1) / (a2@np.cross(a3, a1))
        b3 = 2 * np.pi * np.cross(a1, a2) / (a3@np.cross(a1, a2))
        self.bv: np.ndarray = np.array([b1, b2, b3])

        # Generate atom list
        atom_list = [f'{symbol} '*num
                     for symbol, num in zip(self.atoms, self.natoms)]

        atom_list = ' '.join(atom_list).split()
        self._atom_list = atom_list

        # calculate the volume of the unit cell
        self.V = np.dot(a1, np.cross(a2, a3))

    def CIF_reader(self) -> None:
        '''
        Reads a .cif file and initializes it into the structure class.
        '''
        data = io.read(self.filename) 
        self.name = data.get_chemical_formula()
        self.scale = 1
        self.av = data.cell.array
        self._atom_list = data.get_chemical_symbols()
        self._N = len(self._atom_list)
        self._type = 'direct'
        self.coords = data.get_scaled_positions()

        # convert atom list into atoms and natoms
        atom_list = self._atom_list
        atoms = list(set(atom_list))
        natoms = [atom_list.count(x) for x in atoms]
        self.atoms = atoms
        self.natoms = natoms

        # Calculate reciprocal lattice vectors
        a1, a2, a3 = self.av 
        b1 = 2 * np.pi * np.cross(a2, a3) / (a1@np.cross(a2,a3))
        b2 = 2 * np.pi * np.cross(a3, a1) / (a2@np.cross(a3,a1))
        b3 = 2 * np.pi * np.cross(a1, a2) / (a3@np.cross(a1,a2))
        self.bv: np.ndarray = np.array([b1, b2, b3])

        # calculate the volume of the unit cell
        self.V = np.dot(a1, np.cross(a2, a3))

    @property
    def get_name(self):
        '''
        Getter for name
        '''
        print(f'Name: {self.name}')
        return self.name

    @get_name.setter
    def rename(self, new_name: str):
        '''
        Setter for name
        '''
        print(f'Name changed from "{self.name}" to "{new_name}"')
        self.name = new_name

    @get_name.deleter
    def delete_name(self):
        '''
        Deleter for name
        '''
        print(f'Name "{self.name}" deleted')
        del self.name

    def Cart_coords(self) -> np.ndarray:
        '''
        Converts fractional coordinates to cartesian coordinates.
        '''
        self._type = 'cartesian'
        self.cart_coords = self.coords @ self.av

    def Supercell(self, n: list) -> None:
        '''
        Creates a supercell of size n1 x n2 x n3.
        '''
        coords = self.coords
        av = self.av
        self.orig_av = av
        atom_list = self._atom_list

        _n1 = list(range(n[0]))
        _n2 = list(range(n[1]))
        _n3 = list(range(n[2]))
        _n = np.array([[_n1[i], _n2[j], _n3[k]]
                       for i in range(len(_n1))
                       for j in range(len(_n2))
                       for k in range(len(_n3))])

        coords = np.array([coord + _n[i]
                           for i in range(len(_n))
                           for coord in coords], dtype=float)

        for i in range(len(coords)):
            coords[i, 0] = coords[i, 0] / n[0]
            coords[i, 1] = coords[i, 1] / n[1]
            coords[i, 2] = coords[i, 2] / n[2]

        av[0], av[1], av[2] = av[0] * n[0], av[1] * n[1], av[2] * n[2]
        atom_list = [atom_list] * np.prod(n)
        atom_list = np.array(atom_list).flatten()

        self.coords = coords
        self.av = av
        self._atom_list = atom_list

    def supercell_reorder(self):
        coords = self.coords
        _atom_list = self._atom_list
        coords = np.concatenate((_atom_list.reshape(-1, 1), coords), axis=1)
        coords = coords[coords[:, 0].argsort()]
        self.coords = coords[:, 1:]
        self._atom_list = coords[:, 0]

        atoms = []
        for i in self._atom_list:
            if i not in atoms:
                atoms.append(i)
        
        natoms = [list(self._atom_list).count(i) for i in atoms]

        self.atoms = atoms
        self.natoms = natoms

    def Const_shift(self, vector: np.ndarray) -> np.ndarray:
        '''
        Shifts the structure by a constant vector.
        '''
        coords = self.coords
        coords += vector
        if self._type == 'direct':
            coords %= 1
        else:
            pass
        self.coords = coords

    def Calculate_mesh(self) -> np.ndarray:
        '''
        Calculates K-point mesh.
        '''
        kpoint_mesh = []
        kspacing_min, kspacing_max = 0.05, 0.5
        av = self.av
        bv = self.bv
        bv_norm = np.array([np.linalg.norm(x) for x in bv], dtype=float)

        temp = [(i, norm) for i, norm in enumerate(bv_norm)]
        temp.sort(key=lambda x: x[1], reverse=True)

        i1, i2, i3 = [i for i, _ in temp]

        # Calculate the number of subdivisions N1, N2, N3 in the reciprocal lattice vectors 
        N_max = max(1, int(np.ceil(bv_norm[i1] / kspacing_min)))
        N_min = max(1, int(np.ceil(bv_norm[i1] / kspacing_max)))

        for n1 in range(N_min, N_max):
            min_spacing = bv_norm[i1] / n1
            if np.fabs(bv_norm[i2] - bv_norm[i1]) < 1e-5:
                n2 = n1
            else:
                n2 = int(np.ceil(bv_norm[i2] / min_spacing))
                n2 = max(n2, 1)

            if np.fabs(bv_norm[i3] - bv_norm[i2]) < 1e-5:
                n3 = n2
            else:
                n3 = int(np.ceil(bv_norm[i3] / min_spacing))
                n3 = max(n3, 1)

            if bv_norm[i2] / n2 < kspacing_max and bv_norm[i3] / n3 < kspacing_max:
                mesh = np.array([None, None, None])
            
            mesh[i1], mesh[i2], mesh[i3] = n1, n2, n3 
            kpoint_mesh.append(mesh)

        # calculate kpoint density
        volume = np.linalg.det(bv)
        density = np.array([[np.prod(mesh) / volume] for mesh in kpoint_mesh], dtype=float)

        return np.array(kpoint_mesh, dtype=int), density

    def _ENCUT(self) -> np.ndarray:
        '''
        Calculates possible ENCUT values to test for convergence.
        '''
        # check if POTCAR file exits
        if not os.path.isfile('POTCAR'):
            raise Warning('POTCAR file not found. Please generate POTCAR file first.')

        with open('POTCAR', 'r') as f:
            lines = f.readlines()
        enmax = [float(re.findall(r'ENMAX = ([0-9]+.[0-9]+)', x)[0]) for x in lines if 'ENMAX' in x]
        encut = max(enmax) * 1.3
        encut = round(encut / 50) * 50
        encut_list = [x for x in range(encut - 300, encut + 300, 50)]
        self.encut = encut_list
        return np.array(encut_list)

    def _find_index(self, atom: str):
        '''
        finds the index of the supplied atom in the coords
        and then returns the index of the atom in the atom_list
        '''
        atom = re.split(r'(\d+)', atom)
        for ind, symb in enumerate(self.atoms):
            if symb == atom[0]:
                atom_index = ind

        _prev = np.sum(self.natoms[:atom_index])
        location = int(_prev) + int(atom[1]) - 1
        vector = - self.coords[location]

        self.vector = vector

    def _build_region(self):
        coords = self.cart_coords
        atom_list = self._atom_list
        atom_list = np.array(atom_list).reshape(-1, 1)
        coords = np.concatenate((atom_list, coords), axis=1)
        a1, a2, a3 = self.orig_av
        a1, a2, a3 = np.linalg.norm(a1), np.linalg.norm(a2), np.linalg.norm(a3)
        a = np.min([a1, a2, a3])

        # build a region around the atom of shape a1 x a2 x a3
        region = np.array([coords[i]
                           for i in range(len(coords))
                           if np.linalg.norm(coords[i, 1:]) < a/2])

        coords = np.array(region[:, 1:], dtype=float)
        atom_list = region[:, 0]
        self.cart_coords = coords
        self._atom_list = atom_list
        self.name = self.name + '_qm'

    def write_xyz(self):
        '''
        Writes an xyz file
        '''
        coords = self.cart_coords
        atom_list = self._atom_list
        name = self.name

        with open(f'{name}.xyz', 'w') as f:
            f.write(f'{len(coords)}\n\n')
            for i in range(len(coords)):
                f.write(f'{atom_list[i]} {coords[i,0]} {coords[i,1]} {coords[i,2]}\n')

    def write_poscar(self):
        '''
        Writes POSCAR file. 
        '''
        with open('POSCAR', 'w') as f:
            f.write(f'{self.name}\n')
            f.write(f'{self.scale}\n')
            f.write(f'{" ".join([str(x) for x in self.av[0]])}\n')
            f.write(f'{" ".join([str(x) for x in self.av[1]])}\n')
            f.write(f'{" ".join([str(x) for x in self.av[2]])}\n')
            f.write(f'{" ".join(self.atoms)}\n')
            f.write(f'{" ".join([str(x) for x in self.natoms])}\n')
            f.write(f'{self._type}\n')
            for x in self.coords:
                f.write(f'{" ".join([str(y) for y in x])}\n')


class DOPE():
    """
    A class doping materials.

    Returns:
        POSCAR: POSCAR files containing the doped structures.
    """
    def __init__(
            self, 
            filename: str,
            dopant: str,
            replace: str,
            instances: int):
        '''
        Initializes the DOPE class
        '''
        self.filename = filename
        self.dopant = dopant
        self.replace = replace
        self.instances = instances 
        self._s = Structure(filename)
        self._s.Vasp_reader()

    def Symbol_coords(self) -> np.ndarray: 
        '''
        Returns the symbol and coordinates of the atoms in the structure.
        '''
        return np.hstack((np.array(self._s._atom_list).reshape(-1,1), self._s.coords))

    def _replace_atom(self, coords) -> np.ndarray:
        '''
        Replaces the atom in the structure with the dopant in all possible locations. 
        '''
        dopant = self.dopant 
        replace = self.replace 



        # validate replace atom 
        if replace not in self._s.atoms: 
            raise ValueError(f'Atom {replace} not found in structure.')

        structures = []
        temp_coords = coords.copy()
        prev_index = None
        for i in range(len(coords)):
            if coords[i][0] == replace and i != prev_index:
                temp_coords[i][0] = dopant 
                prev_index = i 
                structures.append(temp_coords) 
                temp_coords = coords.copy()

        return np.array(structures)

    @timer
    def Generate_structure(self) -> np.ndarray: 
        '''
        Generate the doped structures with the correct number of instances.
        '''
        structures = self._replace_atom(self.Symbol_coords()) 
        # transform structures into one array of structures
        instances = self.instances 

        if instances == 1:
            pass
        else:
            for j in range(instances-1): 
                for i in range(len(structures)):
                    _temp = self._replace_atom(structures[i]) 
                    structures = np.vstack((structures, _temp))
                    structures = np.unique(structures, axis=0)

        # remove duplicarte structures 
        structures = np.array([x for x in structures if np.count_nonzero(x[:,0] == self.dopant) == instances]) 
        structures = np.unique(structures, axis=0)
        for i in range(len(structures)):
            structures[i] = structures[i][structures[i][:,0].argsort()]

        self._structures = structures
        print(f'Number of structures found = {len(structures)}')

    def Create_defect(self):
        '''
        Creates defects in the stucture
        '''
        structures = []
        for structure in self._structures:
            for i in range(len(structure)):
                try:
                    if structure[i][0] == self.dopant:
                        structure = np.delete(structure, i, axis=0)
                        structures.append(structure)
                except:
                    pass

        self._structures = np.array(structures)

    def write_poscars(self) -> None:
        '''
        Writes the doped structures to POSCAR files.
        '''
        name: str = self._s.name 
        scale: int  = self._s.scale 
        av: np.ndarray = self._s.av 
        type: str = self._s._type
        structures = self._structures

        for i in range(len(structures)):
            atom_list = structures[i][:,0] 
            atoms, counts = np.unique(atom_list, return_counts=True) 
            counts = np.array([str(x) for x in counts]) 
            coords = structures[i][:,1:]

            with open(f'POSCAR_{i+1}.vasp', 'w') as f:
                f.write(f'{name}\n')
                f.write(f'{scale}\n') 
                f.write(f'{av[0][0]} {av[0][1]} {av[0][2]}\n') 
                f.write(f'{av[1][0]} {av[1][1]} {av[1][2]}\n') 
                f.write(f'{av[2][0]} {av[2][1]} {av[2][2]}\n') 
                f.write(f'{" ".join(atoms)}\n')
                f.write(f'{" ".join(counts)}\n') 
                f.write(f'{type}\n') 
                for coord in coords: 
                    f.write(f'{coord[0]} {coord[1]} {coord[2]}\n')

class Rotate():
    '''
    A class for rotating structures.
    '''
    def __init__(self, filename: str, axis: str, angle: float):
        '''
        Initializes the Rotate class.
        '''
        self.filename = filename 
        self.axis = axis 
        self.angle = angle 
        self._s = Structure(filename) 
        self._s.Vasp_reader()
        self.Rx = np.array([
            [1, 0, 0],
            [0, np.cos(angle), -np.sin(angle)],
            [0, np.sin(angle), np.cos(angle)]
            ])
        self.Ry = np.array([
            [np.cos(angle), 0, np.sin(angle)],
            [0, 1, 0],
            [-np.sin(angle), 0, np.cos(angle)]
            ])
        self.Rz = np.array([
            [np.cos(angle), -np.sin(angle), 0],
            [np.sin(angle), np.cos(angle), 0],
            [0, 0, 1]
            ])

    def translate(self) -> None:
        # find the centre 
        coords = self._s.coords 
        centre = np.mean(coords, axis=0) 
        # translate the structure to the origin 
        self._s.coords = coords - centre 
        self._rotate_coord() 
        self._s.coords = self._s.coords + centre 


    def _rotate_coord(self) -> None:
        '''
        Rotates the structure.
        '''
        # Rotate coordinates
        coords = self._s.coords
        if self.axis == 'x':
            self._s.coords = coords @ self.Rx
            #self.av = self._s.av @ self.Rx
        elif self.axis == 'y':
            self._s.coords = coords @ self.Ry
            #self.av = self._s.av @ self.Ry
        elif self.axis == 'z':
            self._s.coords = coords @ self.Rz
            #self.av = self._s.av @ self.Rz
        else:
            raise ValueError(f'Axis {self.axis} not found.')

    def write_poscar(self) -> None: 
        '''
        Writes the rotated structure to a POSCAR file.
        '''
        name: str = self._s.name 
        scale: int  = self._s.scale 
        av: np.ndarray = self._s.av 
        type: str = self._s._type
        coords = self._s.coords
        counts = [str(x) for x in self._s.natoms]

        with open(f'POSCAR_rotated.vasp', 'w') as f:
            f.write(f'{name}\n')
            f.write(f'{scale}\n') 
            f.write(f'{av[0][0]} {av[0][1]} {av[0][2]}\n') 
            f.write(f'{av[1][0]} {av[1][1]} {av[1][2]}\n') 
            f.write(f'{av[2][0]} {av[2][1]} {av[2][2]}\n') 
            f.write(f'{" ".join(self._s.atoms)}\n')
            f.write(f'{" ".join(counts)}\n')
            f.write(f'{type}\n') 
            for coord in coords: 
                f.write(f'{coord[0]} {coord[1]} {coord[2]}\n')


class Asymmetric_unit():
    '''
    A class for creating asymmetric units from a molecular crystal POSCAR file.
    '''
    def __init__(self,
                 filename: str,
                 atom: str,
                 bond_max: float,
                 ):
        '''
        Initializes the Asymmetric_unit class.
        '''
        self.filename = filename 
        self.atom = atom
        self.bond_max = bond_max
        self._s = Structure(filename) 
        self._s.Vasp_reader() 

    def _locate_atom(self) -> None: 
        '''
        Locates the atom in the structure
        '''
        atom = re.split(r'(\d+)', self.atom)
        symbol, index = atom[0], int(atom[1])

        if symbol not in self._s.atoms:
            raise ValueError(f'Atom {symbol} not found in the structure.') 

        coords = self._s.coords 
        atoms = self._s.atoms 
        natoms = self._s.natoms 

        # find the atom in the structure 
        for ind, sym in enumerate(atoms):
            if sym == symbol:
                atom_index = sum(natoms[0:ind]) + index - 1
        self.vector = np.array([.0, .0, .0]) - coords[atom_index]

    def _translate(self) -> None: 
        '''
        Performs a translation and expansion to ensure a whole molecule is included.
        '''
        self._s.Const_shift(self.vector) 
        self._s.Supercell([2,2,2])
        self._s._type = 'Cartesian'
        self.vector = np.array([-.5, -.5, -.5])
        self._s.Const_shift(self.vector)
        self._s.Cart_coords()

    def _origin_index(self) -> int:
        '''
        Finds the index of the origi atom.
        '''
        coords = self._s.coords
        for ind, val in enumerate(coords):
            if np.allclose(val, [0, 0, 0]):
                return ind

    def _nearest_neighbours(self, coords: np.ndarray, point: np.ndarray, bond_max: float) -> np.ndarray:
        neighbours = sp.spatial.KDTree(coords[:,1:], leafsize=10, compact_nodes=True, balanced_tree=True) 
        dist, ind = neighbours.query(point, k=10)
        ind = [ind[i] for i in range(len(ind)) if dist[i] < bond_max]
        coords = np.array([coords[i] for i in ind])
        return coords

    def _write_xyz(self) -> None:
        '''
        Writes the asymmetric unit to an XYZ file.
        '''
        with open(f'{self._s.name}_asymmetric_unit.xyz', 'w') as f:
            f.write(f'{len(self.asymm_unit)}\n')
            f.write(f'{self._s.name}\n') 
            for atom in self.asymm_unit: 
                f.write(f'{atom[0]} {atom[1]} {atom[2]} {atom[3]}\n')

    @timer
    def find_unit(self) -> None:
        '''
        Finds the asymmetric unit.
        '''
        self._locate_atom()
        self._translate()
        atom_list = np.array(self._s._atom_list).reshape(-1,1)
        origin = self._origin_index()
        coords = np.concatenate((atom_list, self._s.cart_coords), axis=1)

        asymm_unit = np.array([coords[origin]])
        searching = True 
        while searching:
            total_coords = []
            for i in asymm_unit:
                if not i[0] == 'H':
                    temp = self._nearest_neighbours(coords, i[1:], self.bond_max)
                    for i in range(len(temp)): 
                        total_coords.append(temp[i])
            total_coords = np.array(total_coords)
            total_coords = np.unique(total_coords, axis=0)
            if len(total_coords) == len(asymm_unit):
                searching = False
            else: 
                asymm_unit = total_coords
        self.asymm_unit = asymm_unit
        self._write_xyz()



''' A python script to generate input files for the Ewald code '''

# Import modules
from .structure import Structure
import scipy as sp
import numpy as np
import matplotlib.pyplot as plt
from scipy import constants
import re


class Ewald():
    '''
    Calculates the infinite periodic potential of a solid state crystal.
    The system is divided into three zones:
        Zone 1: qm region
        Zone 2: Exact point charge region
        Zone 3: Fitted Parameters

    Parameters
    ----------
    qm_region : file
        xyz file containing the qm region (vasp_suite asymmetric unit)
    filename : str
        name of the POSCAR/ CONTCAR structure file
    charges : file
        file containing the charges of the atoms in the POSCAR
    n : list
        list of sum expansion (2n+1) x (2n+1) x (2n+1)
    r_cut : float
        cutoff radius for the point charge region
    expansion : list
        list of the expansion parameters [ n1, n2, n3 ]
    Atom : str
        atom symbol and index for the central atom of the qm region <Dy1>

    Returns
    -------
    ewald.out : file
        output file containing coordinates and charges for xfield
    '''

    def __init__(
            self,
            qm_region,
            filename,
            charges,
            atom,
            n,
            r_cut,
            expansion,
            bound,
            verbose
            ):
        '''
        Initializes the Ewald class
        '''
        self._s = Structure(filename)
        self._s.Vasp_reader()
        self._charges = charges
        self._n = n
        self._r_cut = r_cut
        self._exp = np.array(expansion)
        self._atom = atom
        self._qm_region = qm_region
        self.read_charges()
        self._s.bv = self._s.bv / (2 * np.pi)
        self.alpha = 0.2
        self.evconv = 1e10 * constants.e / (4 * np.pi * constants.epsilon_0)
        self.bound = bound
        self.verbose = verbose

        # initial checks 
        a1, a2, a3 = self._s.av
        n1, n2, n3 = self._exp
        a1, a2, a3 = a1 * n1, a2 * n2, a3 * n3
        a_v = np.array([a1, a2, a3])
        min = np.linalg.norm(np.min(a_v, axis=0))
        if self._r_cut > min:
            # raise ValueError("Cutoff radius is larger than the smallest lattice vector")
            pass

        # check unit cell charge
        if np.sum(self._q) != 0:
            raise ValueError("Unit cell is not charge neutral")

    def read_charges(self) -> None:
        '''
        Reads the charges from the file
        '''
        with open(self._charges, 'r') as f:
            _q = f.readlines()
        self._q = [float(i) for i in _q]
        if len(self._q) != self._s._N:
            raise ValueError(
                    "Number of charges does not match number of atoms")

    def _uc_print(self) -> None:
        '''
        Creates a print out for the unit cell information
        '''
        print('''
--------------------------------

     Unit Cell Information

--------------------------------
        ''')

        print("Lattice Vectors:")
        for i in range(len(self._s.av)):
            print(f"{self._s.av[i, 0]:.8f} {self._s.av[i, 1]:.8f} {self._s.av[i, 2]:.8f}")
        print("\nCoordinates:")
        for i in range(self._s._N):
            print(f"{self._s.coords[i, 0]:.8f} {self._s.coords[i, 1]:.8f} {self._s.coords[i, 2]:.8f} {self._s._atom_list[i]} {self._q[i]:.8f}")
        print("\nReciprocal Lattice Vectors:\n")
        for i in range(len(self._s.bv)):
            print(f"{self._s.bv[i, 0]:.8f} {self._s.bv[i, 1]:.8f} {self._s.bv[i, 2]:.8f}")

    def _compute_n(self) -> None:
        _n1 = list(range(-self._n[0], self._n[0]+1))
        _n2 = list(range(-self._n[1], self._n[1]+1))
        _n3 = list(range(-self._n[2], self._n[2]+1))
        self._n = np.array([[_n1[i], _n2[j], _n3[k]]
                            for i in range(len(_n1))
                            for j in range(len(_n2))
                            for k in range(len(_n3))], dtype=int)

    def _images(self):
        coords = self._s.coords
        _n = self._n
        _av = self._s.av
        images = np.array([coord + _n[i] for i in range(len(_n))
                           for coord in coords], dtype=float)
        images = images @ _av
        self.images: np.ndarray = images
        self._num_images = len(images)

    def _ewald_recip(self) -> None:
        '''
        Calculates the reciprocal space contribution to the Ewald sum
        '''
        prefactor = (np.pi * self._s.V)**(-1)
        av = self._s.av
        bv = self._s.bv
        q = self._q
        coords = self._s.coords
        _n = self._n

        e_recip = np.zeros((self._s._N, self._s._N), dtype=float)
        f_recip = np.zeros((self._s._N, 3), dtype=float)

        for i in range(self._s._N):
            for j in range(self._s._N):
                for k in range(len(_n)):
                    if i == j or _n[k, 0] == 0 or _n[k, 1] == 0 or _n[k, 2] == 0:
                        continue
                    else:
                        r = coords[i]@av - coords[j]@av
                        m = bv @ _n[k]
                        e_recip[i,j] += q[j] * np.exp(-(np.pi**2 * m@m / self.alpha**2)) / (m@m) * np.cos(2 * np.pi * m @ r) 
                        f_recip[i] += q[j] * np.exp(-(np.pi**2 * m@m / self.alpha**2)) * np.sin(2 * np.pi * m @ r) * m
        self.e_recip = self.evconv * prefactor * e_recip
        self.f_recip = prefactor * f_recip

    def _ewald_real(self) -> None:
        '''
        Calculates the real space contribution to the Ewald sum
        '''
        coords = self._s.coords @ self._s.av
        q = self._q
        q = [q] * len(self._n)
        q = np.array(q, dtype=float).flatten()

        e_real = np.zeros((self._s._N, self._num_images), dtype=float)
        f_real = np.zeros((self._s._N, 3), dtype=float)

        for i in range(self._s._N):
            for j in range(self._num_images):
                if (coords[i] == self.images[j]).all():
                    continue
                else:
                    r = np.linalg.norm(coords[i] - self.images[j])
                    e_real[i, j] = q[j] * sp.special.erfc(self.alpha * r) / r
                    f_real[i] += q[j] * sp.special.erfc(self.alpha * r) * (coords[i] - self.images[j]) / (r**3)
        self.e_real = self.evconv * e_real
        self.f_real = f_real

    def _ewald_self(self):
        e_self = np.zeros(self._s._N, dtype=float)
        f_self = np.zeros((self._s._N, 3), dtype=float)
        q = self._q

        for i in range(self._s._N):
            e_self[i] = -q[i] * 2 * self.alpha / np.sqrt(np.pi)
            f_self[i] = -q[i] * 2 * self.alpha / np.sqrt(np.pi) * self._s.coords[i] @ self._s.av
        self.e_self = self.evconv * e_self
        self.f_self = f_self

    def Calculate_Ewald(self) -> None:
        '''
        Calculates the Ewald sum
        '''
        self._uc_print()
        self._compute_n()
        self._images()
        print('''
--------------------------------

          Ewald Sum

--------------------------------
''')
        self._ewald_recip()
        self._ewald_real()
        self._ewald_self()
        self.e_real, self.e_recip = np.sum(self.e_real, axis=1), np.sum(self.e_recip, axis=1)
        self.e_total = self.e_real + self.e_recip + self.e_self
        for i in range(self.e_total.shape[0]):
            print(f'Index: {i}\nReal: {self.e_real[i]:.5f}, Recip: {self.e_recip[i]:.5f}, Self: {self.e_self[i]:.5f}, Total: {self.e_total[i]:.5f}')

    def _vector(self):
        '''
        Calculates the translation vector for the cell
        '''
        coords = self._s.coords
        atom = self._atom
        atom = re.split(r'(\d+)', atom)
        for ind, sym in enumerate(self._s.atoms):
            if sym == atom[0]:
                atom_index: int = ind

        _prev = np.sum(self._s.natoms[:atom_index])
        location = int(_prev + int(atom[1]) - 1)

        _dest = np.array([
            [0.5, 0.0, 0.0],
            [0.0, 0.5, 0.0],
            [0.0, 0.0, 0.5]
            ], dtype=float)

        if self._exp[0] % 2 == 0:
            _dest[0, 0] = 0.0
        if self._exp[1] % 2 == 0:
            _dest[1, 1] = 0.0
        if self._exp[2] % 2 == 0:
            _dest[2, 2] = 0.0

        _dest = _dest.sum(axis=0)
        _vector = coords[location] - _dest
        coords = coords - _vector
        coords %= 1.0
        self._s.coords = coords

    def _build_supercell(self):
        '''
        Generates the _supercell
        '''
        _exp = self._exp
        _av = self._s.av
        coords = self._s.coords
        e_total = self.e_total
        atom_list = self._s._atom_list

        _n1 = list(range(_exp[0]))
        _n2 = list(range(_exp[1]))
        _n3 = list(range(_exp[2]))

        _n = np.array([[_n1[i], _n2[j], _n3[k]]
                       for i in range(len(_n1))
                       for j in range(len(_n2))
                       for k in range(len(_n3))], dtype=int)

        coords = np.array([coord + _n[i]
                           for i in range(len(_n))
                           for coord in coords], dtype=float)

        for i in range(len(coords)):
            coords[i, 0] = coords[i, 0] / _exp[0] - 0.5
            coords[i, 1] = coords[i, 1] / _exp[1] - 0.5
            coords[i, 2] = coords[i, 2] / _exp[2] - 0.5

        _av[0], _av[1], _av[2] = _av[0] * _exp[0], _av[1] * _exp[1], _av[2] * _exp[2]
        coords = coords @ _av

        e_total = [e_total] * (_exp[0] * _exp[1] * _exp[2])
        e_total = np.array(e_total, dtype=float).flatten()

        q = [self._q] * (_exp[0] * _exp[1] * _exp[2])
        q = np.array(q, dtype=float).flatten()

        atom_list = [atom_list] * (_exp[0] * _exp[1] * _exp[2])
        atom_list = np.array(atom_list, dtype=str).flatten()

        self._supercell: np.ndarray = coords
        self._b = e_total
        self._q = q
        self._atom_list = atom_list

    def _create_zones(self):
        rcut = self._r_cut
        coords = self._supercell
        atom_list = self._atom_list
        q = self._q
        b = self._b
        file = self._qm_region

        with open(file, 'r') as f:
            _qm = f.readlines()
        _qm = [line.strip().split() for line in _qm]
        _qm = np.array(_qm[2:])
        _qm = np.array(_qm[:, 1:], dtype=float)

        _supercell = np.concatenate((coords,
                                     b[:, None],
                                     q[:, None],
                                     atom_list[:, None]), axis=1)

        _sphere = np.array([coord for coord in _supercell
                            if np.linalg.norm(coord[:3]) < rcut])

        _param = np.array([coord for coord in _supercell
                           if np.linalg.norm(coord[:3]) > rcut])

        self._sphere = _sphere
        self._param = _param
        self._qm = _qm
        self._supercell = _supercell

    def _drive_dipole(self) -> float:
        '''
        Calculates the dipole moment and drives it to zero
        '''
        _sphere = self._sphere
        _param = self._param

        ind1, ind2, ind3 = np.random.choice(len(_param), 3, replace=True)

        r = np.array([_param[ind1], _param[ind2], _param[ind3]])

        _r = np.array([_param[ind1, :3],
                       _param[ind2, :3],
                       _param[ind3, :3]], dtype=float)

        q = np.array([_param[ind1, 4],
                      _param[ind2, 4],
                      _param[ind3, 4]], dtype=float)

        # delete the three reandom charges
        np.delete(_param, [ind1, ind2, ind3], axis=0)

        # calculate the dipole moment
        dipole_sphere = np.array(_sphere[:, 4], dtype=float) @ np.array(_sphere[:, :3], dtype=float)
        dipole_param = np.array(_param[:, 4], dtype=float) @ np.array(_param[:, :3], dtype=float)
        dipole_total = dipole_sphere + dipole_param
        print(f'Initial dipole moment = [{dipole_total[0]:.5f} {dipole_total[1]:.5f} {dipole_total[2]:.5f}]')
        d = - np.array(dipole_total, dtype=float)

        # Solve the linear equation system
        x = np.linalg.lstsq(_r.T, d, rcond=None)[0]

        # Modify the charges
        q += x

        dipole_moment = q @ _r
        print(f'Fitted dipole moment = [{dipole_moment[0]:.5f} {dipole_moment[1]:.5f} {dipole_moment[2]:.5f}]')

        # combine the charges and coordinates 
        r[:, 4] = q

        # update the _param 
        _param = np.concatenate((_param, r), axis=0)

        # Calculate the new dipole moment
        dipole_param = np.array(_param[:, 4], dtype=float) @ np.array(_param[:, :3], dtype=float)
        dipole_total = dipole_sphere + dipole_param
        print(f'Final dipole moment = [{dipole_total[0]:.5f} {dipole_total[1]:.5f} {dipole_total[2]:.5f}]') 
        rms = np.sqrt(np.mean(dipole_total**2))
        print(f'RMS = {rms:.5f}\n')
        # Drive the dipole moment to zero
        self._param = _param
        return rms

        # calculate the change in charge for each of the three charges so the total dipole moment is zero 
    def Create_Zones(self) -> None:
        '''
        Creates the zones for the Ewald sum
        '''
        self._vector()
        self._build_supercell()
        self._create_zones()
        print('''
--------------------------------

     Driving Dipole Moment
           To Zero

--------------------------------
Beginning RMS trials...
''')
        rms = 200.0
        run = 0
        while rms > 20:
            rms = self._drive_dipole()
            run += 1
            if run > 100:
                raise ValueError('Could not drive the dipole moment to zero')

    def _fit_params(self):
        verbose: int = self.verbose
        _sphere = np.array(self._sphere)
        _param = np.array(self._param)
        _sphere_coords = np.array(_sphere[:, :3], dtype=float)
        _param_coords = np.array(_param[:, :3], dtype=float)
        print('''
--------------------------------

        Fitting Parameters

--------------------------------
''')

        bi = np.array(_sphere[:, 3], dtype=float)

        A = 1 / sp.spatial.distance.cdist(_sphere_coords, _param_coords, 'euclidean')
        rijs = sp.spatial.distance.squareform(1/sp.spatial.distance.pdist(_sphere_coords, 'euclidean'))
        P = rijs @ (self.evconv * np.array(_sphere[:, 4], dtype=float))
        P_param = A @ (self.evconv * np.array(_param[:, 4], dtype=float))
        bi -= P
        bi -= P_param

        c = np.ones((1, len(_param)))
        d = -(np.sum(np.array(_sphere[:, 4], dtype=float)) + np.sum(np.array(_param[:, 4], dtype=float))) * self.evconv

        Q, R = np.linalg.qr(c.T, mode='complete')

        AQ = A @ Q

        b2 = bi - AQ[:, 0] * (1/R[0]).T * d

        x2 = sp.optimize.lsq_linear(AQ[:, 1:],
                                    b2,
                                    bounds=(-15, 15),
                                    method='bvls',
                                    tol=1e-8,
                                    max_iter=10000,
                                    verbose=verbose)
        x2 = x2.x

        x1 = (1/R[0]).T * d
        x = np.concatenate((x1, x2), axis=0)
        x = Q @ x

        x = x / self.evconv
        x = np.array(_param[:, 4], dtype=float) + x

        # calcualte rmsd
        q_total = np.concatenate((_sphere[:, 4], x), axis=0)
        coords = np.concatenate((_sphere_coords, _param_coords), axis=0)
        rij = sp.spatial.distance.squareform(
                1/sp.spatial.distance.pdist(coords, 'euclidean'))
        rij[np.where(rij == np.inf)] = 0

        potential = self.evconv * np.array(q_total, dtype=float) @ rij
        ewald = np.array(_sphere[:, 3], dtype=float)
        rmsd = np.sqrt(np.sum(
            (potential[:len(_sphere)] - ewald)**2) / len(ewald))

        self.rmsd = rmsd
        self.q_total = q_total
        self.coords = coords
        self._x = x

    def _outputs(self) -> None:
        coords = np.array(np.hstack((self.coords,
                                     self.q_total[:, None])), dtype=str)
        # remove _qm region
        _qm = np.array(self._qm[:, :3], dtype=float)
        for i in range(len(_qm)):
            for j in range(len(coords)-1):
                if np.allclose(_qm[i, :3],
                               np.array(coords[j, :3], dtype=float)):
                    coords = np.delete(coords, j, axis=0)

        with open('ewald.out', 'w') as f:
            for i in range(len(coords)):
                f.write(f'{coords[i,0]} {coords[i,1]} {coords[i,2]} {coords[i,3]}\n')

    def Fitting(self) -> None:
        '''
        Fits the parameters for the Ewald summation
        '''
        self._fit_params()
        print(f'RMSD = {self.rmsd}')
        print(f'Charge of the sphere = {np.sum(np.array(self._sphere[:, 4]), dtype=float)}')
        print(f'Charge of the parameter region = {np.sum(self._x)}')
        print(f'Charge of sphere + parameter region = {np.sum(np.array(self._sphere[:, 4], dtype=float)) + np.sum(self._x)}')
        print(f'Number of sphere charges = {len(self._sphere) - len(self._qm)}')
        print(f'Number of parameter charges = {len(self._param)}')
        print(f'Number of total charges = {len(self._sphere) - len(self._qm) + len(self._param)}')
        self._outputs()


class Potential:
    '''
    Class for plotting potential on a mesh grid
    '''

    def __init__(
            self,
            filename: str,
            xy_grid: int,
            n_points: int,
            n_contours: int,
            ):
        '''
        Initializes the plotter
        '''
        self.filename = filename
        self.xy_grid = xy_grid
        self.n_points = n_points
        self.n_contours = n_contours

    def _data(self) -> None:
        '''
        Reads the data file
        '''
        with open(self.filename, 'r') as f:
            data = f.readlines()
        self.data = np.array([i.strip().split() for i in data], dtype=float)

    def _potential(self) -> None:
        '''
        Construct the mesh grid and calculate the potential
        '''
        data = self.data

        x = np.linspace(-self.xy_grid, self.xy_grid, self.n_points)
        y = np.linspace(-self.xy_grid, self.xy_grid, self.n_points)
        X, Y = np.meshgrid(x, y)

        Z = np.zeros_like(X)
        for i in range(len(X)):
            for j in range(len(X)):
                for k in range(len(data)):
                    Z[i, j] += data[k, 3] / np.sqrt((X[i, j] - data[k, 0])**2 + (Y[i, j] - data[k, 1])**2 + data[k, 2]**2)

        self.X = X
        self.Y = Y
        self.Z = Z

    def Plot(self) -> None:
        '''
        Plots the potential
        '''
        self._data()
        self._potential()

        fig, ax = plt.subplots()
        ax.contourf(self.X, self.Y, self.Z, self.n_contours, cmap='viridis')
        ax.set_xlabel('x/ Å')
        ax.set_ylabel('y/ Å')
        # show colorbar
        cbar = fig.colorbar(ax.contourf(self.X, self.Y, self.Z, self.n_contours, cmap='viridis'))
        cbar.set_label('Potential')
        plt.show()



"""
Handles the automation of phonopy calculations.
"""

import os
import numpy as np
import socket

def disp_files():
    """
    Identifies the number of 'POSCAR-XXX' files in the current directory.
    """
    disp_files = [x for x in os.listdir()
                  if x.startswith('POSCAR-')]

    disp_files = [x.split('-')[1] for x in disp_files]
    disp_files = [int(x) for x in disp_files]

    return max(disp_files)


def check_config():
    """
    Checks if the config file exists.
    """
    if os.path.isfile(os.path.join(os.path.expanduser('~'),
                                   '.vasp_suite_configs/phonon.ini')):
        return True
    else:
        return False


def create_displacements(supercell):
    """
    Uses phonopy to create displacements.
    """

    os.system('phonopy -d --dim="{}"'.format((" ".join(map(str, supercell)))))
    print('Displacements created.')


class phonopy_submit():

    def __init__(self, cores, kpoints):
        self.cores = cores
        self.kpoints = kpoints
        if self.kpoints != [1, 1, 1]:
            self.vasp_type = 'vasp_std'
        else:
            self.vasp_type = 'vasp_gam'
        hostname = socket.gethostname()
        if 'csf3' in hostname:
            self.hostname = 'csf3'
            self.module = 'apps/intel-19.1/vasp/5.4.4'
            self.sub = 'qsub'
        elif 'csf4' in hostname:
            self.hostname = 'csf4'
            self.module = 'vasp/5.4.4-iomkl-2020.02'
            self.sub = 'sbatch'
        else:
            self.hostname = None
            self.sub = 'mpirun ...'

    def write_submit(self, disp):
        with open('submit.sh', 'w') as f:
            if self.hostname == 'csf3':
                f.write("#!/bin/bash --login \n")
                f.write("#$ -cwd \n")
                f.write("#$ -pe smp.pe {} \n".format(self.cores))
                f.write("#$ -N phonopy \n \n")
                f.write("module load {} \n \n".format(self.module))
            if self.hostname == 'csf4':
                f.write("#!/bin/bash --login \n")
                f.write("#SBATCH -p multicore \n")
                f.write("#SBATCH -n {} \n".format(self.cores))
                f.write("#SBATCH --job-name=phonopy \n \n")
                f.write("module load {} \n \n".format(self.module))

            f.write("echo 'Started running at `date`' \n \n")
            f.write("for f in $(ls POSCAR-{001..%03d}) \n" % disp)
            f.write("do \n")
            f.write('if [ -f "${f}" ] \n then \n')
            f.write('d="${f/POSCAR-/}" \n')
            f.write('if [ ! -d "${d}" ] && mkdir "${d}" \n then \n')
            f.write('cd "${d}" \n')
            f.write('cp "../INCAR" "../KPOINTS" "../POTCAR" . \n')
            f.write('mv "../${f}" "POSCAR" \n')
            f.write('mpirum -np {} {}_par '.format(self.cores, self.vasp_type) + '>../${d}.out \n')
            f.write('cd .. \n')
            f.write('fi \n')
            f.write('fi \n')
            f.write('done \n \n')
            f.write("echo 'Finished running at `date`' \n \n")

        os.system('chmod +x submit.sh')
        # print in green
        print('\033[92m' + "{} submit.sh".format("run " + self.sub) + '\033[0m')
